# coding: utf-8

from pathlib import Path
from PIL import Image

from .util import BaseTestInterface, mock_filedialog, mock_warning_window


class Test03LoadImages(BaseTestInterface):

    def testLoadImages(self) -> None:
        """This test checks that MyoFInDer successfully loads several images,
        displays the first one in the image canvas, and displays them all in
        the file table."""

        # The mock selection window returns the paths to three images to load
        mock_filedialog.file_name = [
            str(Path(__file__).parent / 'data' / f'image_{i}.jpg')
            for i in (1, 2, 3)]
        mock_warning_window.WarningWindow.value = 1
        self._window._select_images()

        # Checking that an image is displayed in the image canvas
        self.assertIsInstance(self._window._image_canvas._image, Image.Image)
        self.assertIsInstance(self._window._image_canvas._image_path, Path)
        self.assertEqual(self._window._image_canvas._image_path.name,
                         'image_1.jpg')

        # Checking that the images have entries in the files table
        self.assertEqual(len(self._window._files_table.table_items), 3)
        self.assertEqual(
            self._window._files_table.table_items.current_entry.path.name,
            'image_1.jpg')
        self.assertCountEqual(
            [item.path.name for item in self._window._files_table.table_items],
            [f'image_{i}.jpg' for i in (1, 2, 3)])



# coding: utf-8

from pathlib import Path
from cv2 import (imread, cvtColor, IMREAD_ANYCOLOR, IMREAD_GRAYSCALE,
                 IMREAD_COLOR, COLOR_BGR2RGB)
from numpy import zeros, stack, concatenate, ndarray
from re import findall


def check_image(image_path: Path) -> ndarray | None:
    """Function making sure that the loaded image has 3 channels and is 8-bits.

    It ignores the alpha channel if any, and can handle all the usual dtypes.
    Grayscale images simply receive two additional empty channels to give a
    3-channel image.

    Args:
        image_path: The path to the image to load.

    Returns:
        The loaded image as a 3-channel 8-bits array, or None if the loading
        wasn't successful.
    """

    # Loading the image
    cv_img = imread(str(image_path), IMREAD_ANYCOLOR)

    # In case the file cannot be reached
    if cv_img is None:
        return None

    zero_channel = zeros([cv_img.shape[0], cv_img.shape[1]])

    # In this section, we ensure that the image is RGB
    # The image is grayscale, building a 3 channel image from it
    if len(cv_img.shape) == 2:
        cv_img = stack([concatenate([zero_channel, zero_channel],
                                    axis=2), cv_img], axis=2)

    # If there's an Alpha channel on a grayscale, ignore it
    elif len(cv_img.shape) == 3 and cv_img.shape[2] == 2:
        cv_img = imread(str(image_path), IMREAD_GRAYSCALE)
        cv_img = stack([concatenate([zero_channel, zero_channel],
                                    axis=2), cv_img], axis=2)

    # The image is 3 or 4 channels, ignoring the Alpha channel if any
    elif len(cv_img.shape) == 3 and cv_img.shape[2] in [3, 4]:
        cv_img = imread(str(image_path), IMREAD_COLOR)
        cv_img = cvtColor(cv_img, COLOR_BGR2RGB)

    # If it's another format, returning None to indicate it wasn't successful
    else:
        return None

    # Parsing the d_type string of the image
    try:
        depth = findall(r'\d+', str(cv_img.dtype))[0]
    except IndexError:
        depth = ''
    type_ = str(cv_img.dtype).strip(depth)

    # In this section, we ensure that the image is 8-bits
    # If it's boolean, the image will be black and white
    if type_ == 'bool':
        cv_img = (cv_img.astype('uint8') * 255).astype('uint8')

    # If it's int, first making it uint and then casting to uint8
    elif type_ == 'int':
        cv_img = (cv_img + 2 ** (int(depth) - 1)).astype('uint' +
                                                         depth)
        cv_img = (cv_img / 2 ** (int(depth) - 8)).astype('uint8')

    # If it's uint, simply casting to uint8
    elif type_ == 'uint':
        cv_img = (cv_img / 2 ** (int(depth) - 8)).astype('uint8')

    # If it's float, casting to [0-1] and then to uint8
    elif type_ == 'float':
        cv_img = ((cv_img - cv_img.min()) / (cv_img.max() - cv_img.min())
                  * 255).astype('uint8')

    # If it's another format, returning None to indicate it wasn't successful
    else:
        return None

    return cv_img



# coding: utf-8

from tkinter import TclError

from .util import BaseTestInterface


class Test01Exit(BaseTestInterface):

    def testExit(self) -> None:
        """This test checks that the main window is destroyed as expected after
        a call to safe_destroy()."""

        # Destroying the main window
        self._window.safe_destroy()

        # This call should raise an error as the window shouldn't exist anymore
        with self.assertRaises(TclError):
            self._window.wm_state()

        # Indicating the tearDown() method not to destroy the window
        self._exit = True



# coding: utf-8

from tkinter import Canvas, ttk, Event, messagebox
from PIL import Image, ImageTk, ImageDraw, ImageColor
from platform import system
from functools import partial
from copy import deepcopy
from pathlib import Path
import logging
from typing import Literal

from .tools import Nucleus, Nuclei, Fibers, check_image, SelectionBox


class ImageCanvas(ttk.Frame):
    """This class manages the display of one image, its nuclei and its
    fibers."""

    def __init__(self, mainframe: ttk.Frame, main_window) -> None:
        """Initializes the frame, the layout, the callbacks and the
        variables."""

        # Setting the logger
        self._logger = logging.getLogger("MyoFInDer.ImageCanvas")

        super().__init__(mainframe)

        self.nuclei_table = None
        self._main_window = main_window
        self._settings = self._main_window.settings

        self._set_layout()
        self._set_bindings()

        self.log("Setting the images canvas's variables")

        self._image: Image.Image | None = None
        self._image_path: Path | None = None
        self._img_scale: float = 1.0
        self._can_scale: float = 1.0
        self._delta: float = 1.3
        self._current_zoom: int = 0
        self._nuclei: Nuclei = Nuclei()
        self._fibers: Fibers = Fibers()
        self._image_tk: ImageTk.PhotoImage | None = None
        self._image_id: int | None = None
        self._selection_box: SelectionBox = SelectionBox()

        # Objects used for the display of the fiber overlay
        self._fib_overlay_tk: ImageTk.PhotoImage | None = None
        self._fib_overlay_idx: int | None = None

    def log(self, msg: str) -> None:
        """Wrapper for reducing the verbosity of logging."""

        self._logger.log(logging.INFO, msg)

    def set_indicators(self) -> None:
        """Redraws the nuclei and/or fibers after the user changed the elements
        to display."""

        # Show or hide the nuclei
        if self._settings.show_nuclei.get():
            self.log(f"Showing all the {len(self._nuclei)} nuclei on the "
                     f"canvas")
            for nuc in self._nuclei:
                self._canvas.itemconfigure(nuc.tk_obj, state="normal")
        else:
            self.log(f"Showing all the {len(self._nuclei)} nuclei on the "
                     f"canvas")
            for nuc in self._nuclei:
                self._canvas.itemconfigure(nuc.tk_obj, state="hidden")

        # Nothing more to do if there's no fiber overlay
        if self._fib_overlay_idx is None:
            return

        # Show or hide the fibers
        if self._settings.show_fibers.get():
            self.log(f"Showing all the {len(self._fibers)} fiber contours "
                     f"on the canvas")
            self._canvas.itemconfigure(self._fib_overlay_idx,
                                       state="normal")
        else:
            self.log(f"Hiding all the {len(self._fibers)} fiber contours "
                     f"on the canvas")
            self._canvas.itemconfigure(self._fib_overlay_idx,
                                       state="hidden")

    def load_image(self,
                   path: Path,
                   nuclei: Nuclei,
                   fibers: Fibers) -> None:
        """Loads and displays an image and its fibers and nuclei.

        Args:
            path: The path to the image to load.
            nuclei: The Nuclei object containing the position and color of the
                nuclei to draw on top of the image.
            fibers: The Fibers object containing the position of the fibers to
                draw on top of the image.
        """

        self.log(f"Loading the image {path}")

        # First, checking that the image can be loaded
        # Otherwise, all data would be lost !
        if check_image(path)[0] is None:
            messagebox.showerror(f'Error while loading the image !',
                                 f'Check that the image at '
                                 f'{path} still exists and '
                                 f'that it is accessible.')
            self.log(f"Could not load image {path}, aborting")
            return

        # First, reset the canvas
        self.reset()

        # Then, load and display the image
        self._image_path = path
        self._set_image()
        self.show_image()

        # Deep copy to have independent objects
        self._nuclei = deepcopy(nuclei)

        # Drawing the nuclei
        self.draw_nuclei()

        # Deep copy to have independent objects
        self._fibers = deepcopy(fibers)

        # Drawing the fibers
        self.draw_fibers()

        # Show or hide the fibers depending on the state of the indicators
        self.set_indicators()

    def reset(self) -> None:
        """Resets every object in the canvas: the image, the nuclei and the
        fibers."""

        self.log("Resetting the entire image canvas")

        # Resetting the variables
        self._image = None
        self._image_path = None
        self._img_scale = 1.0
        self._can_scale = 1.0
        self._current_zoom = 0

        # Removing the fibers and nuclei from the canvas
        self.delete_nuclei()
        self.delete_fibers()

        # Resetting the fibers and nuclei objects
        self._nuclei = Nuclei()
        self._fibers = Fibers()

        # Removing the image from the canvas
        if self._image_id is not None:
            self._canvas.delete(self._image_id)
        self._image_id = None
        self._image_tk = None

        # Resetting the selection box
        self._selection_box = SelectionBox()

    @property
    def nuc_col_out(self) -> str:
        """Returns the color of the nuclei outside of fibers, that depends on
        the selected channels.

        The color returned is the RGB color that is neither the one of nuclei
        nor the one of fibers.
        """

        num_to_color = {0: 'blue', 1: '#32CD32', 2: 'red'}
        color_to_num = {'blue': 0, 'green': 1, 'red': 2}

        fiber_num = color_to_num[self._settings.fiber_colour.get()]
        nuclei_num = color_to_num[self._settings.nuclei_colour.get()]

        if not (fiber_num == 2 and nuclei_num == 0):
            return num_to_color[3 - fiber_num - nuclei_num]
        # Special case when the fiber channel is red and the nuclei are blue
        else:
            return 'red'

    @property
    def nuc_col_in(self) -> str:
        """Returns the color of the nuclei inside fibers, that depends on the
        selected channels."""

        if self._settings.nuclei_colour.get() == 'green':
            if self._settings.fiber_colour.get() == 'red':
                return 'magenta'
            else:
                return 'blue'

        elif self._settings.nuclei_colour.get() == 'red':
            if self._settings.fiber_colour.get() == 'blue':
                return '#646464'
            else:
                return 'white'

        return 'yellow'

    @property
    def fib_color(self) -> str:
        """returns the color of the fibers, that depends on the selected
        channels."""

        if self._settings.fiber_colour.get() == 'green':
            return 'magenta'

        elif self._settings.fiber_colour.get() == 'blue':
            if self._settings.nuclei_colour.get() == 'green':
                return 'magenta'
            else:
                return 'white'

        elif self._settings.fiber_colour.get() == 'red':
            if self._settings.nuclei_colour.get() == 'green':
                return 'cyan'
            else:
                return '#FFA500'

        else:
            raise ValueError(f"Got incorrect fiber color: "
                             f"{self._settings.fiber_colour.get()}")

    def show_image(self, *_: Event) -> None:
        """Displays the image on the canvas.

        Args:
            *_: Ignores the event in case the command was issued by one.
        """

        # Making sure an image is available for display
        if self._image is None:
            self.log("Show image called but no image was set, aborting")
            return

        self.log(f"Displaying the image {self._image_path}")

        # Keeping only the channels the user wants
        multiplier = (self._settings.red_channel_bool.get(), 0, 0, 0,
                      0, self._settings.green_channel_bool.get(), 0, 0,
                      0, 0, self._settings.blue_channel_bool.get(), 0)
        image = self._image.convert("RGB", multiplier)

        # Resizing the image to the canvas size
        scaled_x = int(image.width * self._img_scale)
        scaled_y = int(image.height * self._img_scale)
        image = image.resize((scaled_x, scaled_y))

        # Delete previous image on canvas before creating new one
        if self._image_id is not None:
            self._canvas.delete(self._image_id)

        # Actually displaying the image in the canvas
        image_tk = ImageTk.PhotoImage(image)
        self._image_id = self._canvas.create_image(0, 0, anchor='nw',
                                                   image=image_tk)
        self._canvas.lower(self._image_id)
        self._image_tk = image_tk
        self._canvas.configure(scrollregion=(0, 0, scaled_x, scaled_y))

        # Moving the image to the top left corner if it doesn't fill the
        # canvas
        if (scaled_x < self._canvas.winfo_width() or
            scaled_y < self._canvas.winfo_height()):
            self._canvas.xview_moveto(0),
            self._canvas.yview_moveto(0)
    
    def delete_nuclei(self) -> None:
        """Removes all nuclei from the canvas, but doesn't delete the nuclei
        objects."""

        self.log("Deleting all the nuclei on the canvas")

        for nuc in self._nuclei:
            if nuc.tk_obj is not None:
                self._canvas.delete(nuc.tk_obj)
                nuc.tk_obj = None

    def delete_fibers(self) -> None:
        """Removes all fibers from the canvas, but doesn't delete the fibers
         objects."""

        self.log("Deleting all the fibers on the canvas")

        # Delete the overlay object
        if self._fib_overlay_idx is not None:
            self._canvas.delete(self._fib_overlay_idx)

        # Reset the overlay helper objects
        self._fib_overlay_idx = None
        self._fib_overlay_tk = None

    def draw_nuclei(self) -> None:
        """Draws all the nuclei on the canvas, in hidden mode."""

        self.log(f"Drawing all the {len(self._nuclei)} nuclei on the "
                 f"canvas in hidden mode")
        for nuc in self._nuclei:
            nuc.tk_obj = self._draw_nucleus(
                nuc.x_pos, nuc.y_pos,
                self.nuc_col_in if nuc.color == 'in'
                else self.nuc_col_out, 'hidden')

    def draw_fibers(self) -> None:
        """Creates a transparent overlay on which the outline of the detected
        fibers is displayed, and adds it to the canvas in hidden mode.

        This overlay will then be displayed on top of the main image if the
        user enables the fiber indicator.
        """

        if self._image is None:
            return

        self.log(f"Drawing all the {len(self._fibers)} fiber contours "
                 f"on a separate overlay")

        # Draw an empty overlay with the same dimension as the image
        overlay = Image.new("RGBA", (self._image.width, self._image.height),
                            (0, 0, 0, 0))
        draw = ImageDraw.Draw(overlay)

        # Parameters driving fiber aspect
        color = (*ImageColor.getrgb(self.fib_color), 255)
        line_width = max(int(3 * self._can_scale), 1)

        # Actually draw fibers in the overlay
        for fib in self._fibers:
            draw.line(fib.position + [fib.position[0]],
                      fill=color, width=line_width)

        # Get scale factor to adjust overlay to current size
        scaled_x = int(self._image.width * self._img_scale)
        scaled_y = int(self._image.height * self._img_scale)

        # Create the overlay image object
        overlay = overlay.resize((scaled_x, scaled_y))
        overlay_tk = ImageTk.PhotoImage(overlay)

        # Delete the old overlay image to avoid memory leak
        if self._fib_overlay_idx is not None:
            self._canvas.delete(self._fib_overlay_idx)

        # Add the new overlay image to the canvas
        self._fib_overlay_idx = self._canvas.create_image(
            0, 0, anchor="nw", image=overlay_tk, state='hidden')

        # Store reference to keep overlay alive and raise on image
        self._fib_overlay_tk = overlay_tk
        self._canvas.tag_raise(self._fib_overlay_idx)

    def _set_layout(self) -> None:
        """Creates the frame, canvas and scrollbar objects, places them and
        displays them."""

        self.log("Setting the images canvas's layout")

        # Creating the architecture
        self.pack(expand=True, fill="both", anchor="w", side="left",
                  padx=5, pady=5)
        self._hbar_frame = ttk.Frame(self)
        self._hbar_frame.pack(expand=False, fill="x", anchor="s",
                              side="bottom")

        # Creating the canvas
        self._canvas = Canvas(self,
                              highlightthickness=3,
                              highlightbackground="black")
        self._canvas.configure(scrollregion=self._canvas.bbox('all'))
        self._canvas.pack(expand=True, fill="both", side='left')

        # Creating the vertical scrollbar
        self._vbar = ttk.Scrollbar(self, orient="vertical")
        self._vbar.pack(fill='y', side='right')
        self._vbar.config(command=self._canvas.yview)

        # Creating the horizontal scrollbar
        self._hbar = ttk.Scrollbar(self._hbar_frame, orient="horizontal")
        self._hbar.pack(fill='x')
        self._hbar.config(command=self._canvas.xview)

        # Linking the scrollbars to the canvas
        self._canvas.config(yscrollcommand=self._vbar.set)
        self._canvas.config(xscrollcommand=self._hbar.set)
        self._canvas.configure(xscrollincrement='1', yscrollincrement='1')

        # Finally, applying the changes
        self._canvas.update()

    def _set_bindings(self) -> None:
        """Sets the actions to perform for the different user inputs."""

        self.log("Setting the images canvas's bindings")

        self._canvas.bind('<ButtonPress-2>', self._move_from)
        self._canvas.bind('<B2-Motion>', self._move_to)

        # Different mousewheel management in Linux and Windows
        if system() == "Linux":
            self._canvas.bind('<4>', self._zoom)
            self._canvas.bind('<5>', self._zoom)
        else:
            self._canvas.bind('<MouseWheel>', self._zoom)

        # Moving with the arrow keys
        self.bind_all('<Left>', self._arrows)
        self.bind_all('<Right>', self._arrows)
        self.bind_all('<Up>', self._arrows)
        self.bind_all('<Down>', self._arrows)

        # Zooming in and out with keyboard keys
        self.bind_all('=', partial(self._zoom, delta=1, mouse=False))
        self.bind_all('+', partial(self._zoom, delta=1, mouse=False))
        self.bind_all('-', partial(self._zoom, delta=-1, mouse=False))
        self.bind_all('_', partial(self._zoom, delta=-1, mouse=False))

        # Clicking with the mouse
        self._canvas.bind('<ButtonRelease-1>', self._left_release)
        self._canvas.bind('<ButtonRelease-3>', self._right_release)
        self._canvas.bind('<ButtonPress-1>', self._click)
        self._canvas.bind('<ButtonPress-3>', self._click)
        self._canvas.bind('<B1-Motion>', self._motion)
        self._canvas.bind('<B3-Motion>', self._motion)

    def _draw_nucleus(self,
                      unscaled_x: float,
                      unscaled_y: float,
                      color: str,
                      state: Literal['normal', 'hidden']) -> int:
        """Draws a single nucleus on the canvas.

        Args:
            unscaled_x: The x position of the nucleus on the raw image, in
                pixels.
            unscaled_y: The y position of the nucleus on the raw image, in
                pixels.
            color: The color of the nucleus.
            state: Whether the nucleus should be displayed right away or not.

        Returns:
            The index tkinter attributed to the nucleus it just drew.
        """

        # Adjusting the radius to the scale
        radius = max(int(3.0 * self._can_scale), 1)

        # Adjusting the position to the scale
        x = unscaled_x * self._img_scale
        y = unscaled_y * self._img_scale

        # Actually drawing the nucleus
        return self._canvas.create_oval(
            x - radius, y - radius, x + radius, y + radius,
            fill=color, outline='#fff', width=0, state=state)

    def _draw_box(self) -> int:
        """Draws the selection box for either inverting the nuclei or deleting
        them."""

        # Adjusting the width of the line to the scale
        line_width = max(int(3 * self._can_scale), 1)

        # Creating the rectangle
        return self._canvas.create_rectangle(
            (self._selection_box.x_start * self._img_scale,
             self._selection_box.y_start * self._img_scale,
             self._selection_box.x_end * self._img_scale,
             self._selection_box.y_end * self._img_scale),
            fill='', outline='#888', width=line_width)

    def _click(self, event: Event) -> None:
        """Binding method for a left or right click event.

        Checks whether the click is inside the displayed image, and if so
        stores its coordinates for potentially displaying later a selection
        box.

        Args:
            event: The tkinter event associated with the left or right click.
        """

        # Do nothing if there's no image displayed
        if self._image is None:
            return

        # The coordinates of the click on the unscaled image
        abs_x = self._canvas.canvasx(event.x) / self._img_scale
        abs_y = self._canvas.canvasy(event.y) / self._img_scale

        # Do something only if the click is inside the image
        if abs_x >= self._image.width or abs_y >= self._image.height:
            return

        # Saving the position of the click
        self._selection_box.x_start = abs_x
        self._selection_box.y_start = abs_y

    def _motion(self, event: Event) -> None:
        """Binding method for a mouse motion event while either the left or
        the right click is being pressed.

        If all the conditions are met, is simply draws or updates the selection
        box.

        Args:
            event: The tkinter event associated with the mouse motion.
        """

        # Do nothing if there's no image displayed
        # Also nothing if the selection box wasn't initialized on first click
        if self._image is None or not self._selection_box.started:
            return

        # The coordinates of the event on the unscaled image
        abs_x = self._canvas.canvasx(event.x) / self._img_scale
        abs_y = self._canvas.canvasy(event.y) / self._img_scale

        # Do something only if the event is inside the image
        if abs_x >= self._image.width or abs_y >= self._image.height:
            return

        # Saving the end position of the selection box
        self._selection_box.x_end = abs_x
        self._selection_box.y_end = abs_y

        # Drawing the selection box if it is not already displayed
        if self._selection_box.tk_obj is None:
            self._selection_box.tk_obj = self._draw_box()

        # Otherwise just updating the displayed selection box
        else:
            self._canvas.coords(
                self._selection_box.tk_obj,
                (self._selection_box.x_start * self._img_scale,
                 self._selection_box.y_start * self._img_scale,
                 self._selection_box.x_end * self._img_scale,
                 self._selection_box.y_end * self._img_scale))

    def _left_release(self, event: Event) -> None:
        """Binding method for a left mouse button release event.

        It simply decides if this event should be handled as a single click or
        if a selection box was drawn.

        Args:
            event: The tkinter event associated with the left mouse button
                release.
        """

        # Selection box is only considered if it exists and if it's big enough
        if (self._selection_box and
            self._selection_box.area * self._img_scale ** 2 > 25):
            self._left_release_box(event)
        # Otherwise the event is treated as a sole click
        else:
            self._left_release_nucleus(event)

        # Cleans up all the selection box related objects before returning
        if self._selection_box.tk_obj is not None:
            self._canvas.delete(self._selection_box.tk_obj)
        self._selection_box = SelectionBox()

    def _left_release_box(self, event: Event) -> None:
        """Method called when the user releases the left mouse button after
        drawing a selection box.

        It inverts all the nuclei found inside the selection box.

        Args:
            event: The tkinter event associated with the left mouse button
                release.
        """

        # Do nothing if there's no image displayed
        if self._image is None:
            return

        # Do nothing if the nuclei are not being displayed
        if not self._settings.show_nuclei.get():
            return

        # The coordinates of the click on the unscaled image
        abs_x = self._canvas.canvasx(event.x) / self._img_scale
        abs_y = self._canvas.canvasy(event.y) / self._img_scale

        # Update coordinates only if the click is inside the image
        if abs_x < self._image.width and abs_y < self._image.height:
            self._selection_box.x_end = abs_x
            self._selection_box.y_end = abs_y

        self.log(f"Inversion box drawn from ("
                 f"{self._selection_box.x_start}, "
                 f"{self._selection_box.y_start}) to ("
                 f"{self._selection_box.x_end}, "
                 f"{self._selection_box.y_end})")

        # Inverting all the nuclei found inside the selection box
        for nuc in self._nuclei:
            if self._selection_box.is_inside(nuc.x_pos, nuc.y_pos):
                self._invert_nucleus(nuc)

    def _left_release_nucleus(self, event: Event) -> None:
        """Upon left click, either adds a new nucleus or a fiber, or switches
        the color of an existing nucleus.

        Args:
            event: The tkinter event associated with the left click.
        """

        # Do nothing if there's no image displayed
        if self._image is None:
            return

        # Do nothing if the nuclei are not being displayed
        if not self._settings.show_nuclei.get():
            return

        # The coordinates of the click on the unscaled image
        abs_x = self._canvas.canvasx(event.x) / self._img_scale
        abs_y = self._canvas.canvasy(event.y) / self._img_scale

        # Do something only if the event is inside the image
        if abs_x >= self._image.width or abs_y >= self._image.height:
            return

        # Trying to find a close nucleus
        nuc = self._find_closest_nucleus(abs_x, abs_y)

        # One close nucleus found, inverting it color
        if nuc is not None:
            self._invert_nucleus(nuc)

        # No close nucleus found, adding a new one
        else:
            self._add_nucleus(abs_x, abs_y)

    def _right_release(self, event: Event) -> None:
        """Binding method for a right mouse button release event.

        It simply decides if this event should be handled as a single click or
        if a selection box was drawn.

        Args:
            event: The tkinter event associated with the right mouse button
                release.
        """

        # Selection box is only considered if it exists and if it's big enough
        if (self._selection_box and
            self._selection_box.area * self._img_scale ** 2 > 25):
            self._right_release_box(event)
        # Otherwise the event is treated as a sole click
        else:
            self._right_release_nucleus(event)

        # Cleans up all the selection box related objects before returning
        if self._selection_box.tk_obj is not None:
            self._canvas.delete(self._selection_box.tk_obj)
        self._selection_box = SelectionBox()

    def _right_release_box(self, event: Event) -> None:
        """Method called when the user releases the right mouse button after
        drawing a selection box.

        It deletes all the nuclei found inside the selection box.

        Args:
            event: The tkinter event associated with the right mouse button
                release.
        """

        # Do nothing if there's no image displayed
        if self._image is None:
            return

        # Do nothing if the nuclei are not being displayed
        if not self._settings.show_nuclei.get():
            return

        # The coordinates of the click on the unscaled image
        abs_x = self._canvas.canvasx(event.x) / self._img_scale
        abs_y = self._canvas.canvasy(event.y) / self._img_scale

        # Update coordinates only if the click is inside the image
        if abs_x < self._image.width and abs_y < self._image.height:
            self._selection_box.x_end = abs_x
            self._selection_box.y_end = abs_y

        self.log(f"Deletion box drawn from ("
                 f"{self._selection_box.x_start}, "
                 f"{self._selection_box.y_start}) to ("
                 f"{self._selection_box.x_end}, "
                 f"{self._selection_box.y_end})")

        to_delete = list()
        # Detecting all the nuclei to delete
        for nuc in self._nuclei:
            if self._selection_box.is_inside(nuc.x_pos, nuc.y_pos):
                to_delete.append(nuc)
        # Actually deleting the detected nuclei
        for nuc in to_delete:
            self._delete_nucleus(nuc)

    def _right_release_nucleus(self, event: Event) -> None:
        """Method called when the user releases the right mouse button but did
        not draw a selection box.

        Tries to find a nucleus close to the click, and if one was found
        deletes it.

        Args:
            event: The tkinter event associated with the right mouse button
                release.
        """

        # Do nothing if there's no image displayed
        if self._image is None:
            return

        # Do nothing if the nuclei are not being displayed
        if not self._settings.show_nuclei.get():
            return

        # The coordinates of the click on the unscaled image
        abs_x = self._canvas.canvasx(event.x) / self._img_scale
        abs_y = self._canvas.canvasy(event.y) / self._img_scale

        # Do something only if the click is inside the image
        if abs_x >= self._image.width or abs_y >= self._image.height:
            return

        # Trying to find a close nucleus
        nuc = self._find_closest_nucleus(abs_x, abs_y)

        # One close nucleus found, deleting it
        if nuc is not None:
            self._delete_nucleus(nuc)

    def _find_closest_nucleus(self, x: float, y: float) -> Nucleus | None:
        """Searches for a close nucleus among the existing nuclei.

        Args:
            x: The x position where to search for a nucleus.
            y: The y position where to search for a nucleus.

        Returns:
            The Nucleus object closest to the given coordinates, if any close
            enough nucleus was found.
        """

        # Adjusting the search radius to the scale
        radius = max(3.0 * self._can_scale / self._img_scale, 3.0)

        closest_nuc = None
        closest_distance = None

        # Iterating over the nuclei
        for nuc in self._nuclei:
            # Keeping only those within the radius of search
            if abs(x - nuc.x_pos) <= radius and abs(y - nuc.y_pos) <= radius:
                dis_sq = (nuc.x_pos - x) ** 2 + (nuc.y_pos - y) ** 2
                # Keeping only the closest one
                if closest_distance is None or dis_sq < closest_distance:
                    closest_distance = dis_sq
                    closest_nuc = nuc

        return closest_nuc if closest_nuc is not None else None

    def _add_nucleus(self, abs_x: float, abs_y: float) -> None:
        """Creates a Nucleus object, displays it and saves it to the nuclei
        table.

        Also sets the unsaved status for the current project.

        Args:
            abs_x: The x position of the nucleus on the unscaled image
            abs_y: The y position of the nucleus on the unscaled image
        """

        # Creating the nucleus and displaying it
        new_nuc = Nucleus(abs_x, abs_y,
                          self._draw_nucleus(abs_x, abs_y,
                                             self.nuc_col_out,
                                             'normal'),
                          'out')

        self.log(f"Added nucleus at position ({abs_x}, {abs_y})")

        # Adding the nucleus to the nuclei table and to the current nuclei
        self.nuclei_table.add_nucleus(new_nuc)
        self._nuclei.append(new_nuc)

        # Setting the unsaved status
        self._main_window.set_unsaved_status()

    def _invert_nucleus(self, nuc: Nucleus) -> None:
        """Inverts the color and status of a given nucleus.

        Also sets the unsaved status for the current project.

        Args:
            nuc: The Nucleus object to invert
        """

        # Updating the display
        if nuc.color == 'out':
            self._canvas.itemconfig(nuc.tk_obj, fill=self.nuc_col_in)
            nuc.color = 'in'
        else:
            self._canvas.itemconfig(nuc.tk_obj, fill=self.nuc_col_out)
            nuc.color = 'out'

        self.log(f"Inverted nucleus at position ({nuc.x_pos}, {nuc.y_pos})")

        # Updating the nuclei table
        self.nuclei_table.switch_nucleus(nuc)

        # Setting the unsaved status
        self._main_window.set_unsaved_status()

    def _delete_nucleus(self, nuc: Nucleus) -> None:
        """Deletes a given Nucleus and all the references to it.

        Also sets the unsaved status for the current project.

        Args:
            nuc: The Nucleus object to delete
        """

        # Deleting all the references to the nucleus to delete
        self.nuclei_table.remove_nucleus(nuc)
        self._canvas.delete(nuc.tk_obj)
        self._nuclei.remove(nuc)

        self.log(f"Deleted nucleus at position ({nuc.x_pos}, {nuc.y_pos})")

        # Setting the unsaved status
        self._main_window.set_unsaved_status()

    def _set_image(self) -> None:
        """Simply loads an image and resizes it to the current size of the
        canvas."""

        if not self._image_path:
            self.log("Image loading requested but no file was selected, "
                     "ignoring")
            return

        self.log(f"Loading image {self._image_path}")

        # Loading the image
        cv_img = check_image(self._image_path)
        # Handling the case when the image cannot be loaded
        if cv_img is None:
            messagebox.showerror(f'Error while loading the image !',
                                 f'Check that the image at '
                                 f'{self._image_path} still exists and '
                                 f'that it is accessible.')
            self.log(f"ERROR ! Could not load image {self._image_path}")
            return

        self._image = Image.fromarray(cv_img)

        # Getting the different parameters of interest
        can_width = self._canvas.winfo_width()
        can_height = self._canvas.winfo_height()
        can_ratio = can_width / can_height
        img_ratio = self._image.width / self._image.height

        # Calculating the new image width
        if img_ratio >= can_ratio:
            resize_width = can_width
        else:
            resize_width = int(can_height * img_ratio)

        # Deducing the new image scale
        self._img_scale = resize_width / self._image.width

    def _arrows(self, event: Event) -> None:
        """Scrolls the image upon pressing on the arrow keys.

        Args:
            event: The event associated with the arrow press.
        """

        self._canvas.configure(xscrollincrement='10', yscrollincrement='10')

        if event.keysym == 'Left':
            self._canvas.xview_scroll(-1, "units")
        elif event.keysym == 'Right':
            self._canvas.xview_scroll(1, "units")
        elif event.keysym == 'Up':
            self._canvas.yview_scroll(-1, "units")
        elif event.keysym == 'Down':
            self._canvas.yview_scroll(1, "units")

        self._canvas.configure(xscrollincrement='1', yscrollincrement='1')

    def _move_from(self, event: Event) -> None:
        """Stores the previous coordinates for scrolling with the mouse."""

        if self._image is not None:
            self._canvas.scan_mark(event.x, event.y)

    def _move_to(self, event: Event) -> None:
        """Drags the canvas to a new position."""

        if self._image is not None:
            self._canvas.scan_dragto(event.x, event.y, gain=1)

    def _zoom(self,
              event: Event,
              delta: int | None = None,
              mouse: bool = True) -> None:
        """Zooms in or out of the image.

        Args:
            event: The tkinter event associated with the scrolling command.
            delta: If positive, zooms in, if negative, zooms out.
            mouse: Does the zoom command come from the mouse wheel ? If yes,
                zoom to the current position of the mouse, else to the top left
                corner of the image.
        """

        if self._image is None:
            return

        # Linux mousewheel events do not hold information about the delta,
        # so we need to set it manually
        if delta is None:
            if system() == "Linux":
                delta = 1 if event.num == 4 else -1
            else:
                delta = int(event.delta / abs(event.delta))

        scale = 1.0
        # Setting the point where to zoom on the image
        x_can = self._canvas.canvasx(event.x) if mouse else 0
        y_can = self._canvas.canvasy(event.y) if mouse else 0

        # Zooming out
        if delta < 0:
            # Limit to 4 zooms out
            if self._current_zoom < -4:
                return

            # Setting the new image and canvas scales
            self._img_scale /= self._delta
            self._can_scale /= self._delta
            scale /= self._delta
            self._current_zoom -= 1

            # Scrolling the canvas to keep the mouse on the same point
            self._canvas.xview_scroll(int((1 / self._delta - 1) * x_can),
                                      "units")
            self._canvas.yview_scroll(int((1 / self._delta - 1) * y_can),
                                      "units")

        # Zooming in
        elif delta > 0:
            # Limit to 4 zooms in
            if self._current_zoom > 4:
                return

            # Setting the new image and canvas scales
            self._img_scale *= self._delta
            self._can_scale *= self._delta
            scale *= self._delta
            self._current_zoom += 1

        # Rescaling the canvas
        self._canvas.scale('all', 0, 0, scale, scale)
        # Updating the image
        self.show_image()

        # Scrolling the canvas to keep the mouse on the same point in case
        # we were zooming in and the image is bigger than the canvas
        if delta > 0 and self._current_zoom > 0:
            self._canvas.xview_scroll(int((self._delta - 1) * x_can),
                                      "units")
            self._canvas.yview_scroll(int((self._delta - 1) * y_can),
                                      "units")



# coding: utf-8

from pathlib import Path
from threading import Thread
from time import sleep

from .util import (BaseTestInterfaceProcessing, mock_filedialog,
                   mock_warning_window)


class Test15NewProject(BaseTestInterfaceProcessing):

    def testNewProject(self) -> None:
        """This test checks that creating a new empty project erases all the
        previous modifications."""

        # The mock selection window returns the path to one image to load
        mock_filedialog.file_name = [
            str(Path(__file__).parent / 'data' / 'image_1.jpg')]
        mock_warning_window.WarningWindow.value = 1
        # The images will be saved to the existing temporary folder created
        # before starting the test
        mock_filedialog.save_folder = str(Path(self._dir.name) / 'save_folder')
        self._window._select_images()

        # Stopping the regular processing Thread
        self._window._stop_thread = True
        sleep(2)
        # Instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)

        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()

        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Starting the processing loop in the main Thread rather than in a
        # separate one
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Triggering a save action using the "Save" button
        self._window._save_button.invoke()

        # Overriding the current changes with a new empty project
        index = self._window._file_menu.index("New Empty Project")
        self._window._file_menu.invoke(index)

        # Checking that the image canvas and the files table are now empty
        self.assertIsNone(self._window._image_canvas._image)
        self.assertIsNone(self._window._image_canvas._image_path)
        self.assertEqual(len(self._window._files_table.table_items), 0)



# coding: utf-8

from pathlib import Path
from threading import Thread
from time import sleep

from .util import (BaseTestInterfaceProcessing, mock_filedialog,
                   mock_warning_window)


class Test11ProcessImages(BaseTestInterfaceProcessing):

    def testProcessImages(self) -> None:
        """This test checks that images can be successfully processed by
        MyoFInDer, and the partial selection checkboxes in the files table also
        work for partial processing."""

        # The mock selection window returns the paths to three images to load
        mock_filedialog.file_name = [
            str(Path(__file__).parent / 'data' / f'image_{i}.jpg')
            for i in (1, 2, 3)]
        mock_warning_window.WarningWindow.value = 1
        self._window._select_images()

        # Stopping the regular processing Thread
        self._window._stop_thread = True
        sleep(2)
        # Instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)

        # Checking that the image canvas is empty of nuclei
        self.assertEqual(len(self._window._image_canvas._nuclei), 0)
        self.assertEqual(self._window._image_canvas._nuclei.nuclei_in_count,
                         0)
        self.assertEqual(self._window._image_canvas._nuclei.nuclei_out_count,
                         0)
        # Checking that the files table is empty of nuclei and fibers
        for i in range(3):
            self.assertEqual(
                len(self._window._files_table.table_items.entries[i].nuclei),
                0)
            self.assertEqual((self._window._files_table.table_items.entries[i].
                              nuclei.nuclei_in_count), 0)
            self.assertEqual((self._window._files_table.table_items.entries[i].
                              nuclei.nuclei_out_count), 0)
            self.assertEqual(
                len(self._window._files_table.table_items.entries[i].fibers),
                0)
            self.assertEqual((self._window._files_table.table_items.entries[i].
                              fibers.area), 0)

        # Invoking the selection checkbox of the last image in the files table
        # to unselect it
        (self._window._files_table.table_items.entries[-1].graph_elt.
         check_button.invoke())
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()

        # Starting the Thread for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Starting the processing loop in the main Thread rather than in a
        # separate one
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that both positive and negative nuclei have been detected on
        # the image canvas
        self.assertGreater(len(self._window._image_canvas._nuclei), 0)
        self.assertGreater(self._window._image_canvas._nuclei.nuclei_in_count,
                           0)
        self.assertGreater(self._window._image_canvas._nuclei.nuclei_out_count,
                           0)
        # Checking that the two first images have their nuclei and fiber counts
        # updated in the files table
        for i in range(2):
            self.assertGreater(
                len(self._window._files_table.table_items.entries[i].nuclei),
                0)
            self.assertGreater((self._window._files_table.table_items.
                                entries[i].nuclei.nuclei_in_count), 0)
            self.assertGreater((self._window._files_table.table_items.
                                entries[i].nuclei.nuclei_out_count), 0)
            self.assertGreater(
                len(self._window._files_table.table_items.entries[i].fibers),
                0)
            self.assertGreater((self._window._files_table.table_items.
                                entries[i].fibers.area), 0)

        # Checking that the nuclei and fibers count of the last image remained
        # unchanged in the files table
        self.assertEqual(
            len(self._window._files_table.table_items.entries[2].nuclei),
            0)
        self.assertEqual((self._window._files_table.table_items.entries[2].
                          nuclei.nuclei_in_count), 0)
        self.assertEqual((self._window._files_table.table_items.entries[2].
                          nuclei.nuclei_out_count), 0)
        self.assertEqual(
            len(self._window._files_table.table_items.entries[2].fibers),
            0)
        self.assertEqual((self._window._files_table.table_items.entries[2].
                          fibers.area), 0)



# coding: utf-8

from tkinter import ttk, Canvas, messagebox, Event, Frame
from xlsxwriter import Workbook
from shutil import copyfile, rmtree
from cv2 import polylines, ellipse, imwrite, cvtColor, COLOR_RGB2BGR
from pathlib import Path
from platform import system
from copy import deepcopy
from functools import partial
from pickle import dump, load
from numpy import ndarray, array
import logging

from .tools import (Nucleus, Fiber, Nuclei, Fibers, GraphicalElement,
                    check_image, TableItems, TableEntry)


class FilesTable(ttk.Frame):
    """This frame stores and displays the different images of the project.

    It also stores the nuclei and fiber count, and handles the saving of images
    and data.
    """

    def __init__(self,
                 root: ttk.Frame,
                 main_window) -> None:
        """Sets the appearance of the frame.

        Args:
            root: The parent frame in which this one is included.
            main_window: the main window of the GUI.
        """

        # Setting the logger
        self._logger = logging.getLogger("MyoFInDer.FilesTable")

        super().__init__(root)

        self._main_window = main_window

        # Setting the appearance
        self._set_layout()

        self.log("Setting the file table's bindings")
        self._canvas.bind('<Configure>', self._on_resize)

        self.image_canvas = None
        self.table_items = TableItems()

    def log(self, msg: str) -> None:
        """Wrapper for reducing the verbosity of logging."""

        self._logger.log(logging.INFO, msg)

    def load_project(self, directory: Path) -> None:
        """Loads an existing project.

        Gets the images names and paths, the fibers positions, and the nuclei
        positions and colors from the data.pickle file.

        Args:
            directory: The path to the directory where the project to load is.
        """

        # First, resetting the project
        self.log("Resetting all the table items")
        self.table_items.reset()

        # Loading the simple version of the stored data
        with open(directory / 'data.pickle', 'rb') as save_file:
            self.log(f"Loading project data from {directory / 'data.pickle'}")
            table_items = load(save_file)

        # Keeping only the data corresponding to existing images
        self.table_items = TableItems(
            entries=[entry for entry in table_items if
                     (directory / 'Original Images' / entry.path).is_file()])

        # Completing the paths to match that of the directory to load
        for entry in self.table_items:
            entry.path = directory / 'Original Images' / entry.path
        loaded = (entry.path for entry in self.table_items)
        self.log(f"Loaded data for images: {', '.join(map(str, loaded))}")

        # Removing any reference to Tkinter objects
        for entry in self.table_items:
            for nucleus in entry.nuclei:
                nucleus.tk_obj = None

        # Redrawing the canvas
        if self.table_items:
            self._make_table()

    def save_project(self, directory: Path, save_overlay: bool) -> None:
        """Saves a project.

        Saves the images names and paths, the fibers positions, the nuclei
        positions and colors, the images themselves, a .xlsx file containing
        stats of the project, and optionally the images with nuclei and fibers
        drawn on it.

        Args:
            directory: The path to the directory where the project has to be
                saved.
            save_overlay: Should the images with nuclei and fibers drawn be
                saved ?
        """

        self.log(f"Saving the project {directory}")

        # First, save the stats in a .xlsx file
        self._save_table(directory)

        # Then, save the original images
        self._save_originals(directory)

        # If needed, save the overlay images
        if save_overlay:
            self._save_overlay_images(directory)

        # Finally, save the data
        self._save_data(directory)

    def add_nucleus(self, nucleus: Nucleus) -> None:
        """Adds a Nucleus to the Nuclei object associated with the current
        image.

        Args:
            nucleus: The Nucleus object to add.
        """

        self.log(f"Nucleus added at position {nucleus.x_pos}, {nucleus.y_pos}")
        self.table_items.current_entry.nuclei.append(deepcopy(nucleus))
        self._update_data(self.table_items.current_entry)

    def remove_nucleus(self, nucleus: Nucleus) -> None:
        """Removes a Nucleus from the Nuclei object associated with the current
        image.

        Args:
            nucleus: The Nucleus object to remove.
        """

        self.log(f"Nucleus removed at position "
                 f"{nucleus.x_pos}, {nucleus.y_pos}")
        self.table_items.current_entry.nuclei.remove(nucleus)
        self._update_data(self.table_items.current_entry)

    def switch_nucleus(self, nucleus: Nucleus) -> None:
        """Switches the color of a nucleus.

        Args:
            nucleus: The Nucleus object whose color should be switched.
        """

        self.log(f"Nucleus switched at position "
                 f"{nucleus.x_pos}, {nucleus.y_pos}")
        for nuc in self.table_items.current_entry.nuclei:
            if nuc == nucleus:
                nuc.color = 'out' if nuc.color == 'in' else 'in'
        self._update_data(self.table_items.current_entry)

    def add_images(self, filenames: list[Path]) -> None:
        """Adds images to the current frame.

        The new images are added under the existing list of images.

        Args:
            filenames: A list containing the paths to the images to load.
        """

        # Deleting every element in the canvas
        self.log("Deleting all the elements in the image canvas")
        self.table_items.reset_graphics()

        # Making sure there's no conflict between the new and previous images
        for file in filenames:
            if file in self.table_items.file_names:
                messagebox.showerror("Error while loading files",
                                     f"The file {file.name} is already opened,"
                                     f"ignoring.")
                self.log(f"Removing file {file} from the images to add as it "
                         f"is already opened")
                filenames.remove(file)
            elif file.name in [file_.name for file_
                               in self.table_items.file_names]:
                messagebox.showerror("Error while  loading files",
                                     f"A file with the same name ({file.name})"
                                     f"is already opened, ignoring.")
                self.log(f"Removing file {file} from the images to add as an "
                         f"image with the same name is already opened")
                filenames.remove(file)

        self.log(f"Adding the images {', '.join(map(str, filenames))}")

        # Adding the new images to the frame
        for file in filenames:
            self.table_items.append(TableEntry(path=file,
                                               nuclei=Nuclei(),
                                               fibers=Fibers()))

        # Redrawing the canvas
        self._make_table()

    def input_processed_data(
            self,
            nuclei_negative_positions: list[tuple[ndarray, ndarray]],
            nuclei_positive_positions: list[tuple[ndarray, ndarray]],
            fiber_contours: tuple[ndarray, ...],
            area: float,
            file: Path) -> None:
        """Adds the nuclei and fibers positions to the frame after they've been
        computed.

        Args:
            nuclei_negative_positions: The positions of the nuclei outside the
                fibers.
            nuclei_positive_positions: The positions of the nuclei inside the
                fibers.
            fiber_contours: The positions of the contour points of the fibers.
            area: The ratio of fiber area over the total image area.
            file: The path to the file whose nuclei and fibers have been
                computed.
        """

        # Adds the received nuclei to the Nuclei object and draws them
        self.table_items[file].nuclei.reset()
        for x, y in nuclei_negative_positions:
            self.table_items[file].nuclei.append(Nucleus(x, y, None, 'out'))
        for x, y in nuclei_positive_positions:
            self.table_items[file].nuclei.append(Nucleus(x, y, None, 'in'))

        # Adds the received fibers to the Fibers object and draws them
        self.table_items[file].fibers.reset()
        self.table_items[file].fibers.area = area
        for contour in fiber_contours:
            # Ensuring single points are not being outlined
            if len(contour.shape) < 2:
                continue
            contour = list(map(tuple, contour))
            self.table_items[file].fibers.append(Fiber(contour))

        # Updates the display
        self._update_data(self.table_items[file])

        # Refreshes the image display if necessary
        if file == self.table_items.selected:
            self.image_canvas.load_image(
                self.table_items.current_entry.path,
                self.table_items.current_entry.nuclei,
                self.table_items.current_entry.fibers)

    def enable_buttons(self, enable: bool) -> None:
        """Enables or disables the close buttons and checkboxes of the images.

        They should be disabled when the program is computing, as the images
        shouldn't be removed at this moment.

        Args:
            enable: Whether the buttons should be enabled or disabled.
        """

        if enable:
            self.log("Enabling the buttons")
        else:
            self.log("Disabling the buttons")

        state = 'normal' if enable else 'disabled'

        for entry in self.table_items:
            entry.graph_elt.check_button['state'] = state
            entry.graph_elt.close_button['state'] = state

    @property
    def all_checked(self) -> bool:
        """Returns True when all the checkboxes of the images are set."""

        # If there's no image, default is True
        if not self.table_items:
            return True

        return all(entry.graph_elt.button_var.get() for entry
                   in self.table_items)

    @property
    def checked_files(self) -> list[Path]:
        """Returns the paths to the files whose checkboxes are checked."""

        return [entry.path for entry in self.table_items
                if entry.graph_elt.button_var.get()]

    def delete_image(self, to_delete: tuple[Path, ...]) -> None:
        """Removes images from the canvas, and discards all the associated
        data.

        Also handles the update of the index of the image currently displayed.

        Args:
            to_delete: A tuple containing the paths to the images to remove
                from the canvas.
        """

        self.log("Image deletion requested by the user")

        # Security to prevent unwanted deletions
        if len(to_delete) == 1:
            ret = messagebox.askyesno('Hold on !',
                                      f"Do you really want to remove "
                                      f"{to_delete[0].name} ?\n"
                                      f"All the unsaved data will be lost.")

        else:
            ret = messagebox.askyesno('Hold on !',
                                      f"Do you really want to remove "
                                      f"{len(to_delete)} files ?\n"
                                      f"All the unsaved data will be lost.")
        if not ret:
            self.log("Image deletion canceled by the user")
            return

        # Cleaning up the canvas
        self.log("Deleting all the elements in the image canvas")
        self.table_items.reset_graphics()

        # Iterating over the files to delete
        for file in to_delete:

            self.log(f"Deleting the file: {file}")

            # Removing any reference to the image being deleted
            index = self.table_items.index(file)
            self.table_items.remove(file)

            # If there's no entry left in the canvas, resetting the image
            if not self.table_items:
                self.table_items.current_index = None
                self.image_canvas.reset()

            # Selecting a new current image if the prev one was just deleted
            elif self.table_items.current_index == index:
                if len(self.table_items) <= index:
                    self.table_items.current_index = len(self.table_items) - 1

            # Updating the current index if a lower index image was deleted
            elif self.table_items.current_index > index:
                self.table_items.current_index -= 1

        # Re-drawing the canvas
        self._make_table()

        # Updating the master checkbox
        self._main_window.update_master_check()

    def _save_data(self, directory: Path) -> None:
        """Saves the names of the images, the positions of the fibers and the
        positions and colors of the nuclei.

        All this information is stored in a data.pickle file.

        Args:
            directory: The directory where the file should be saved.
        """

        with open(directory / 'data.pickle', 'wb+') as save_file:
            self.log(f"Saving the data to file {directory / 'data.pickle'}")
            dump(self.table_items.save_version, save_file, protocol=4)

    def _save_originals(self, directory: Path) -> None:
        """Saves the original images in a sub-folder of the project folder.

        Args:
            directory: The path to the project folder.
        """

        # Creating the directory if it doesn't exit
        if not (directory / 'Original Images').is_dir():
            self.log(f"Creating the folder for saving the original images at: "
                     f"{directory / 'Original Images'}")
            Path.mkdir(directory / 'Original Images')

        self.log("Saving the original images")

        # Actually saving the images
        for item in self.table_items:
            # Saving only if the images are not saved yet
            if directory not in item.path.parents:
                new_path = directory / 'Original Images' / item.path.name
                self.log(f"Saving the image: {new_path}")

                # Handling the case when the image cannot be loaded
                try:
                    copyfile(item.path, new_path)
                except FileNotFoundError:
                    messagebox.showerror(f'Error while saving the image !',
                                         f'Check that the image at '
                                         f'{item.path} still exists and '
                                         f'that it is accessible.')
                    self.log(f"ERROR! Could not save file {new_path}")
                    continue

                # Changing the saved path for the image
                item.path = new_path

            else:
                self.log(f"Skipping {item.path} as it is already saved")

    def _save_table(self, directory: Path) -> None:
        """Saves a .xlsx file containing stats about the images of the project.

        Args:
            directory: The path to the project folder.
        """

        # Creating the Excel file
        workbook = Workbook(str(directory / str(directory.name + '.xlsx')))
        worksheet = workbook.add_worksheet()

        self.log(f"Saving an overview of the data as an Excel file at "
                 f"{directory / str(directory.name + '.xlsx')}")

        # Bold style for a nicer layout
        bold = workbook.add_format({'bold': True, 'align': 'center'})

        # Writing the labels
        worksheet.write(0, 0, "Image names", bold)
        worksheet.write(0, 1, "Total number of nuclei", bold)
        worksheet.write(0, 2, "Number of tropomyosin positive nuclei", bold)
        worksheet.write(0, 3, "Fusion index", bold)
        worksheet.write(0, 4, "Fiber area ratio", bold)

        # Setting the column widths
        worksheet.set_column(
            0, 0, width=max(11, max(len(file.name) for file
                                    in self.table_items.file_names)
                            if self.table_items else 0))
        worksheet.set_column(1, 1, width=22)
        worksheet.set_column(2, 2, width=37)
        worksheet.set_column(3, 3, width=17)
        worksheet.set_column(4, 4, width=16)

        for i, item in enumerate(self.table_items):
            # Writing the names of the images
            worksheet.write(i + 2, 0, item.path.name)

            # Writing the total number of nuclei
            worksheet.write(i + 2, 1, len(item.nuclei))

            # Writing the number of nuclei in fibers
            worksheet.write(i + 2, 2, item.nuclei.nuclei_in_count)

            # Writing the ratio of nuclei in over the total number of nuclei
            if item.nuclei.nuclei_out_count > 0:
                worksheet.write(i + 2, 3, item.nuclei.nuclei_in_count /
                                len(item.nuclei))
            else:
                worksheet.write(i + 2, 3, 'NA')

            # Writing the percentage area of fibers
            worksheet.write(i + 2, 4, f'{item.fibers.area * 100:.2f}')

        workbook.close()

    def _save_overlay_images(self, directory: Path) -> None:
        """Saves the images with the nuclei and fibers drawn on them.

        Args:
            directory: The path to the project folder.
        """

        # Creates the directory if it doesn't exist yet
        if (directory / 'Overlay Images').is_dir():
            rmtree(directory / 'Overlay Images')
            self.log(f"Creating the folder for saving the overlay images at: "
                     f"{directory / 'Overlay Images'}")
        Path.mkdir(directory / 'Overlay Images')

        self.log("Saving the overlay images")

        # Saves the images
        for entry in self.table_items:
            self._draw_nuclei_save(entry, directory)

    def _draw_nuclei_save(self,
                          entry: TableEntry,
                          project_name: Path) -> None:
        """Draws fibers and nuclei on an images and then saves it.

        Args:
            entry: The table entry whose image to save.
            project_name: The path to the project folder.
        """

        color_to_bgr = {'blue': (255, 0, 0),
                        'green': (0, 255, 0),
                        'red': (0, 0, 255),
                        'yellow': (0, 255, 255),
                        'cyan': (255, 255, 0),
                        'magenta': (255, 0, 255),
                        'black': (0, 0, 0),
                        'white': (255, 255, 255),
                        '#FFA500': (0, 165, 255),
                        '#32CD32': (50, 205, 50),
                        '#646464': (100, 100, 100)}

        # Reads the image
        cv_img = check_image(project_name / "Original Images" /
                             entry.path.name)

        destination = project_name / "Overlay Images" / entry.path.name

        # Aborting if the image cannot be loaded
        if cv_img is None:
            self.log(f"Skipping the image {destination} as the original image "
                     f"could not be loaded")
            return

        cv_img = cvtColor(cv_img, COLOR_RGB2BGR)

        # Adjusting the indicator sizes to the size of the image
        max_dim = max(cv_img.shape)
        line_width = max(1, round(max_dim / 1080))
        spot_size = max(1, round(max_dim / 1080 * 2))

        # Drawing the fibers
        for fib in entry.fibers:
            positions = array(fib.position)
            positions = positions.reshape((-1, 1, 2))
            polylines(cv_img, [positions], True,
                      color_to_bgr[self.image_canvas.fib_color], line_width)

        # Drawing the nuclei
        for nuc in entry.nuclei:
            centre = (int(nuc.x_pos), int(nuc.y_pos))
            if nuc.color == 'out':
                ellipse(cv_img, centre, (spot_size, spot_size), 0, 0, 360,
                        color_to_bgr[self.image_canvas.nuc_col_out], -1)
            else:
                ellipse(cv_img, centre, (spot_size, spot_size), 0, 0, 360,
                        color_to_bgr[self.image_canvas.nuc_col_in], -1)

        # Now saving the image
        self.log(f"Saving the image {destination}")
        imwrite(str(destination), cv_img)

    def _set_layout(self) -> None:
        """Sets the layout of the frame by creating the canvas and the
        scrollbar."""

        self.log("Setting the file table's layout")

        self.pack(expand=True, fill="both", anchor="n", side='top')

        # Creating a frame to put a black border around the canvas
        self._canvas_frame = Frame(self,
                                   highlightbackground='black',
                                   highlightthickness=3)
        self._canvas_frame.pack(expand=True, fill="both", anchor="w",
                                side='left', pady=5)

        # Creating the canvas containing the opened images
        self._canvas = Canvas(self._canvas_frame, bg='#FFFFFF')
        self._canvas.configure(scrollregion=self._canvas.bbox('all'))
        self._canvas.pack(expand=True, fill="both", anchor="w", side='left')

        # Creating a frame inside the canvas that will contain all the
        # individual file frames
        self._scroll_window = Frame(self._canvas, bg='#FFFFFF')
        self._id = self._canvas.create_window(0, 0,
                                              window=self._scroll_window,
                                              anchor='nw')

        # Creating the scrollbar
        self._vbar = ttk.Scrollbar(self, orient="vertical")
        self._vbar.pack(side='right', fill='y', pady=5)
        self._vbar.config(command=self._canvas.yview)
        self._canvas.config(yscrollcommand=self._vbar.set)

        self.update()

    def _on_resize(self, event: Event) -> None:
        """Updates the canvas scroll region and the entries width when the
        overall size of the canvas changes.

        Args:
            event: The resize event.
        """

        # When first drawing the canvas an event of width 0 is received
        width = event.width if event.width > 0 else self._canvas.winfo_width()

        # To make the next lines easier to read
        req_height = self._scroll_window.winfo_reqheight()
        can_height = self._canvas.winfo_height()

        # Event of height 0 received and entries do not fill the canvas
        if event.height == 0 and req_height <= can_height:
            height = can_height
        # Event of height 0 received and entries do not fit in the canvas
        elif event.height == 0:
            height = req_height
        # Entries do not fill the canvas
        elif req_height <= event.height:
            height = event.height
        # Entries do not fit in the canvas
        else:
            height = req_height

        # Re-dimensioning the scroll region and the canvas for a nice layout
        self._canvas.config(scrollregion=(0, 0, width, height))
        self._canvas.itemconfig(self._id, width=width)

    def _on_wheel(self, event: Event) -> None:
        """Scrolls the canvas up or down upon wheel motion."""

        # Different wheel management in Windows and Linux
        if system() == "Linux":
            delta = 1 if event.num == 4 else -1
        else:
            delta = int(event.delta / abs(event.delta))

        if self._scroll_window.winfo_reqheight() > self._canvas.winfo_height():
            self._canvas.yview_scroll(-delta, "units")

    def _select_image(self, _: Event, index: int) -> None:
        """Unselects the previous entry, selects the new one and displays the
        associated image and data.

        Args:
            _: The mouse click event that triggered the image selection.
            index: The index in the canvas of the entry that was just clicked.
        """

        self.log(f"Entry with index {index} selected by the user")

        # Getting the entry at the given index
        entry = self.table_items.entries[index]

        # If the entry is already the selected one, do nothing
        if entry.graph_elt.selected.get():
            self.log(f"Skipping as it is already the current selected one")
            return

        # Unselecting the previously selected entry
        if self.table_items.selected is not None:
            self.table_items.current_entry.graph_elt.selected.set(False)

        # Selecting the new entry
        entry.graph_elt.selected.set(True)
        self.table_items.current_index = index

        # Displaying the selected image with its nuclei and fibers
        self.image_canvas.load_image(entry.path, entry.nuclei, entry.fibers)

        self.log(f"Loaded data for the entry {entry.path}")

    def _make_table(self) -> None:
        """Draws the entire canvas."""

        self.log("Drawing the entire files table")

        # The canvas is built iteratively
        for i, entry in enumerate(self.table_items):

            # Setting the file name
            file_name = entry.path.name
            if len(file_name) >= 39:
                file_name = f'...{file_name[-36:]}'

            # Creating a graphical element representing the entry
            entry.graph_elt = GraphicalElement(
                canvas=self._scroll_window,
                number=i,
                name=file_name,
                delete_cmd=partial(self.delete_image, (entry.path,)),
                check_cmd=self._main_window.update_master_check,
                scroll_cmd=self._on_wheel,
                select_cmd=self._select_image)

            # Updating the displayed text according to the data
            self._update_data(entry)

        # Highlighting the selected image
        if self.table_items:
            # Keeping the current index if it exists
            if self.table_items.current_index is not None:
                index = self.table_items.current_index
            # Setting the current index to 0 otherwise
            else:
                index = 0
            # Sending an event for displaying the selected image
            self.table_items.entries[index].graph_elt.event_generate(
                '<ButtonPress-1>', when="tail")

        # Updating the canvas and making sure it configures itself as requested
        self.update()
        self._canvas.event_generate("<Configure>", when="tail")

    def _update_data(self, entry: TableEntry) -> None:
        """Updates the display when data about an image has changed.

        Args:
            entry: The path to the image whose data should be updated.
        """

        self.log(f"Updating the data for the entry {entry.path}")

        # To avoid repeated calls to the same attribute
        nuclei = entry.nuclei
        fibers = entry.fibers
        items = entry.graph_elt
        total_nuclei = len(nuclei)

        # Updating the labels
        items.total.configure(text=f'Total : {total_nuclei}')
        items.positive.configure(text=f'Positive : {nuclei.nuclei_in_count}')
        if total_nuclei > 0:
            items.ratio.configure(
                text=f'Ratio : '
                     f'{int(100 * nuclei.nuclei_in_count / total_nuclei)}%')
        else:
            items.ratio.configure(text='Ratio : NA')
        items.area.configure(text=f'Fiber area : {int(fibers.area * 100)}%')



# coding: utf-8

from tkinter import Toplevel, ttk, Tk
from screeninfo import get_monitors
from pathlib import Path


class SavePopup(Toplevel):
    """Popup window displayed while the project is being saved to inform the
    user that saving is in progress."""

    def __init__(self, main_window: Tk, directory: Path) -> None:
        """Sets the layout of the popup window.

        Args:
            main_window: The parent window of which this popup depends.
            directory: The directory where the project is being saved.
        """

        super().__init__(main_window)

        # Setting the layout
        self.title("Saving....")
        ttk.Label(self, text=f"Saving to '{directory.name}' ..."). \
            pack(anchor='center', expand=False, fill='none', padx=10, pady=10)

        # Centering on the screen
        self.update()
        self._center()

    def _center(self) -> None:
        """Centers the popup window on the currently used monitor."""

        # Getting the current monitor
        monitors = get_monitors()
        candidates = [monitor for monitor in monitors if
                      0 <= self.winfo_x() - monitor.x <= monitor.width and
                      0 <= self.winfo_y() - monitor.y <= monitor.height]
        current_monitor = candidates[0] if candidates else monitors[0]

        # Getting the parameters of interest
        scr_width = current_monitor.width
        scr_height = current_monitor.height
        x_offset = current_monitor.x
        y_offset = current_monitor.y
        height = self.winfo_height()
        width = self.winfo_width()

        # Actually centering the window
        self.geometry('+%d+%d' % (x_offset + (scr_width - width) / 2,
                                  y_offset + (scr_height - height) / 2))



# coding: utf-8

from copy import deepcopy
from pathlib import Path
from threading import Thread
from time import sleep

from .util import (BaseTestInterfaceProcessing, mock_filedialog,
                   mock_warning_window)


class Test19ProcessVarySettings(BaseTestInterfaceProcessing):

    def testProcessVarySettings(self) -> None:
        """This test checks that the output of the processing is influenced as
        expected by the value of some settings."""

        # Copying the initial state of the settings for later restoring it
        init_settings = deepcopy(self._window.settings.get_all())

        # The mock selection window returns the path to one image to load
        mock_filedialog.file_name = [
            str(Path(__file__).parent / 'data' / 'image_1.jpg')]
        mock_warning_window.WarningWindow.value = 1
        # The images will be saved to the existing temporary folder created
        # before starting the test
        mock_filedialog.save_folder = str(Path(self._dir.name) / 'save_folder')
        self._window._select_images()

        # Stopping the regular processing Thread
        self._window._stop_thread = True
        sleep(2)
        # Instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)

        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()

        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Starting the processing loop in the main Thread rather than in a
        # separate one
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Recording the result of the first computation for later comparing it
        nuc = len(self._window._files_table.table_items.entries[0].nuclei)
        nuc_in = (self._window._files_table.table_items.entries[0].
                  nuclei.nuclei_in_count)
        nuc_out = (self._window._files_table.table_items.entries[0].
                   nuclei.nuclei_out_count)
        fib = len(self._window._files_table.table_items.entries[0].fibers)
        area = self._window._files_table.table_items.entries[0].fibers.area

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the minimum fiber intensity setting to the maximum
        self._window._settings_window._min_fib_int_slider.set(
            self._window._settings_window._min_fib_int_slider.cget('to'))

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that the correct computation output changed and the correct
        # ones were preserved
        self.assertEqual(
            nuc, len(self._window._files_table.table_items.entries[0].nuclei))
        self.assertNotEqual(
            nuc_in, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_in_count)
        self.assertNotEqual(
            nuc_out, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_out_count)
        self.assertNotEqual(
            fib, len(self._window._files_table.table_items.entries[0].fibers))
        self.assertNotEqual(
            area, self._window._files_table.table_items.entries[0].fibers.area)

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the maximum fiber intensity setting to the minimum
        self._window._settings_window._max_fib_int_slider.set(
            self._window._settings_window._max_fib_int_slider.cget('from'))

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that the correct computation output changed and the correct
        # ones were preserved
        self.assertEqual(
            nuc, len(self._window._files_table.table_items.entries[0].nuclei))
        self.assertNotEqual(
            nuc_in, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_in_count)
        self.assertNotEqual(
            nuc_out, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_out_count)
        self.assertNotEqual(
            fib, len(self._window._files_table.table_items.entries[0].fibers))
        self.assertNotEqual(
            area, self._window._files_table.table_items.entries[0].fibers.area)

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the minimum nucleus intensity setting to the maximum
        self._window._settings_window._min_nuc_int_slider.set(
            self._window._settings_window._min_nuc_int_slider.cget('to'))

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that the correct computation output changed and the correct
        # ones were preserved
        self.assertNotEqual(
            nuc, len(self._window._files_table.table_items.entries[0].nuclei))
        self.assertNotEqual(
            nuc_in, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_in_count)
        self.assertNotEqual(
            nuc_out, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_out_count)
        self.assertEqual(
            fib, len(self._window._files_table.table_items.entries[0].fibers))
        self.assertEqual(
            area, self._window._files_table.table_items.entries[0].fibers.area)

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the maximum nucleus intensity setting to the minimum
        self._window._settings_window._max_nuc_int_slider.set(
            self._window._settings_window._max_nuc_int_slider.cget('from'))

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that the correct computation output changed and the correct
        # ones were preserved
        self.assertNotEqual(
            nuc, len(self._window._files_table.table_items.entries[0].nuclei))
        self.assertNotEqual(
            nuc_in, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_in_count)
        self.assertNotEqual(
            nuc_out, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_out_count)
        self.assertEqual(
            fib, len(self._window._files_table.table_items.entries[0].fibers))
        self.assertEqual(
            area, self._window._files_table.table_items.entries[0].fibers.area)

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the minimum nucleus diameter setting to the maximum
        self._window._settings_window._min_nuc_diam_slider.set(
            self._window._settings_window._min_nuc_diam_slider.cget('to'))

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that the correct computation output changed and the correct
        # ones were preserved
        self.assertNotEqual(
            nuc, len(self._window._files_table.table_items.entries[0].nuclei))
        self.assertNotEqual(
            nuc_in, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_in_count)
        self.assertNotEqual(
            nuc_out, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_out_count)
        self.assertEqual(
            fib, len(self._window._files_table.table_items.entries[0].fibers))
        self.assertEqual(
            area, self._window._files_table.table_items.entries[0].fibers.area)

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the minimum fiber intensity setting to 50
        self._window._settings_window._min_fib_int_slider.set(50)

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Setting a new standard specifically for the last test
        nuc = len(self._window._files_table.table_items.entries[0].nuclei)
        nuc_in = (self._window._files_table.table_items.entries[0].
                  nuclei.nuclei_in_count)
        nuc_out = (self._window._files_table.table_items.entries[0].
                   nuclei.nuclei_out_count)
        fib = len(self._window._files_table.table_items.entries[0].fibers)
        area = self._window._files_table.table_items.entries[0].fibers.area

        # Restoring the initial state of the settings
        self._window.settings.update(init_settings)

        # Invoking the creation of a settings window
        index = self._window._settings_menu.index("Settings")
        self._window._settings_menu.invoke(index)

        # Setting the minimum nuclei count setting to the maximum
        self._window._settings_window._count_slider.set(
            self._window._settings_window._count_slider.cget('to'))
        # Also setting the fiber threshold to 50, otherwise there is no effect
        self._window._settings_window._min_fib_int_slider.set(50)

        # Closing the settings window
        self._window._settings_window.destroy()

        # Re-instantiating a Thread that will stop the processing loop later on
        stop_thread = Thread(target=self._stop_thread)
        # Invoking the button for starting the image processing
        self._window._process_images_button.invoke()
        # Starting the Threads for later stopping the processing loop
        self._window._stop_thread = False
        stop_thread.start()
        # Re-starting the processing loop in the main Thread
        self._window._process_thread()
        self._window._handle_ui_queue()
        self._window._handle_ui_queue()

        # Checking that the correct computation output changed and the correct
        # ones were preserved
        self.assertEqual(
            nuc, len(self._window._files_table.table_items.entries[0].nuclei))
        self.assertNotEqual(
            nuc_in, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_in_count)
        self.assertNotEqual(
            nuc_out, self._window._files_table.table_items.entries[0].
            nuclei.nuclei_out_count)
        self.assertEqual(
            fib, len(self._window._files_table.table_items.entries[0].fibers))
        self.assertEqual(
            area, self._window._files_table.table_items.entries[0].fibers.area)



# coding: utf-8

import cellpose.models
import numpy as np
import numpy.ma as ma
from pathlib import Path
import cv2
from typing import Any
from collections import defaultdict

from .tools import check_image

# Table for converting color strings to channels, assuming RGB images
numpy_color_to_int = {"red": 0,
                      "green": 1,
                      "blue": 2}


class ImageSegmentation:
    """Class for processing images, detecting fibers and nuclei."""

    def __init__(self) -> None:
        """Simply loads the Mesmer library."""

        self._app = cellpose.models.CellposeModel(diam_mean=17,
                                                  model_type='nuclei')

    def __call__(self,
                 path: Path,
                 nuclei_color: str,
                 fiber_color: str,
                 minimum_fiber_intensity: int,
                 maximum_fiber_intensity: int,
                 minimum_nucleus_intensity: int,
                 maximum_nucleus_intensity: int,
                 minimum_nucleus_diameter: int,
                 minimum_nuclei_count: int
                 ) -> tuple[Path, list[tuple[np.ndarray, np.ndarray]],
                            list[tuple[np.ndarray, np.ndarray]],
                            tuple[Any, ...], float]:
        """Computes the nuclei positions and optionally the fibers positions.

        Also returns whether the nuclei are inside or outside the fibers.

        Args:
            path: The path to the image to process.
            nuclei_color: The color of the nuclei, as a string.
            fiber_color: The color of the fibers, as a string.
            minimum_fiber_intensity: The gray level intensity above which a
                pixel is considered to be part of a fiber.
            maximum_fiber_intensity: The gray level intensity below which a
                pixel is considered to be part of a fiber.
            minimum_nucleus_intensity: Any nucleus whose average brightness is
                lower than this value will be discarded.
            maximum_nucleus_intensity: Any nucleus whose average brightness is
                greater than this value will be discarded.
            minimum_nucleus_diameter: Objects whose area is lower than this
                value (in pixels) will not be considered.
            minimum_nuclei_count: Nuclei located in a fiber containing less
                than this number of positive nuclei will be counted as
                negative.

        Returns:
            The list of nuclei outside the fibers, the list of nuclei inside
            the fibers, the list of fiber contours, and the ratio of fiber area
            over the total area.
        """

        # Converting colors from string to int
        colors = [numpy_color_to_int[nuclei_color],
                  numpy_color_to_int[fiber_color]]

        # Loads the image and keeps only the nuclei and fibers channels
        image = check_image(path)

        # The image couldn't be loaded
        if image is None:
            raise IOError("Could not load the image for segmentation, "
                          "aborting !")

        nuclei_channel = image[:, :, colors[0]]
        fiber_channel = image[:, :, colors[1]]

        del image

        small_objects_threshold = int(minimum_nucleus_diameter ** 2
                                      * np.pi / 4)

        # Image to pass to CellPose
        if nuclei_channel.max() > nuclei_channel.min():
            x = np.stack(((nuclei_channel - nuclei_channel.min()) /
                          (nuclei_channel.max() - nuclei_channel.min()),
                          np.zeros_like(nuclei_channel)),
                         axis=-1)
        else:
            x = np.stack((np.full_like(nuclei_channel, 1.0),
                          np.zeros_like(nuclei_channel)),
                         axis=-1)

        # Actual nuclei detection function
        labeled_image, *_ = self._app.eval(
            # Normalize here instead of letting eval do it
            x=x,
            batch_size=8,
            resample=None,
            channels=None,
            channel_axis=2,
            z_axis=None,
            normalize=False,
            invert=False,
            rescale=None,
            diameter=None,
            flow_threshold=0.4,
            cellprob_threshold=0.0,
            do_3D=False,
            anisotropy=None,
            flow3D_smooth=0,
            stitch_threshold=0.0,
            min_size=small_objects_threshold,
            max_size_fraction=1.0,
            niter=None,
            augment=False,
            tile_overlap=0.1,
            bsize=224,
            compute_masks=True,
            progress=None)

        # Getting the fiber mask
        mask = self._get_fiber_mask(fiber_channel,
                                    nuclei_channel,
                                    minimum_fiber_intensity,
                                    maximum_fiber_intensity)

        del fiber_channel

        # Calculating the area of fibers over the total area
        area = np.count_nonzero(mask) / mask.shape[0] / mask.shape[1]

        # Finding the contours of the fibers
        mask_8_bits = (mask * 255).astype('uint8')
        fiber_contours, _ = cv2.findContours(mask_8_bits, cv2.RETR_LIST,
                                             cv2.CHAIN_APPROX_SIMPLE)
        fiber_contours = tuple(map(np.squeeze, fiber_contours))

        # Getting the position of the nuclei
        nuclei_out, nuclei_in = self._get_nuclei_positions(
            labeled_image, mask, nuclei_channel, 0.75,
            minimum_nucleus_intensity, maximum_nucleus_intensity,
            minimum_nuclei_count)

        return path, nuclei_out, nuclei_in, fiber_contours, area

    @staticmethod
    def _get_fiber_mask(fiber_channel: np.ndarray,
                        nuclei_channel: np.ndarray,
                        minimum_intensity: int,
                        maximum_intensity: int) -> np.ndarray:
        """Applies several images processing methods to the fiber channel of
        the image to smoothen the outline.

        Args:
            fiber_channel: The channel of the image containing the fibers.
            nuclei_channel: The channel of the image containing the nuclei.
            minimum_intensity: The gray level intensity above which a pixel is
                considered to be part of a fiber.
            minimum_intensity: The gray level intensity below which a pixel is
                considered to be part of a fiber.

        Returns:
            A boolean mask containing the position of the fibers
        """

        # Apply a Gaussian filter to smoothen the fiber signal
        dim = max(int(min(fiber_channel.shape) / 50) // 2 * 2 - 1, 1)
        fiber_channel = cv2.GaussianBlur(fiber_channel, (dim, dim), 0)

        # First, apply a base threshold
        kernel = np.ones((4, 4), np.uint8)
        _, min_thresh = cv2.threshold(fiber_channel, minimum_intensity, 255,
                                      cv2.THRESH_BINARY)
        _, max_thresh = cv2.threshold(fiber_channel, maximum_intensity, 255,
                                      cv2.THRESH_BINARY)
        processed = min_thresh - max_thresh
        del min_thresh, max_thresh

        # Opening, to remove noise in the background
        processed = cv2.morphologyEx(processed,
                                     cv2.MORPH_OPEN, kernel)

        # Closing, to remove noise inside the fibers
        processed = cv2.morphologyEx(processed, cv2.MORPH_CLOSE, kernel)

        # Find contours and hierarchy of the holes inside the fibers
        contours, hierarchy = cv2.findContours(processed, cv2.RETR_TREE,
                                               cv2.CHAIN_APPROX_SIMPLE)

        if hierarchy is None:
            return np.zeros_like(processed, dtype=np.bool)

        hierarchy = np.squeeze(hierarchy).tolist()

        # If the hierarchy only contains one object, it is cast to int
        if isinstance(hierarchy, int):
            hierarchy = [hierarchy]
        if (isinstance(hierarchy, list)
            and len(hierarchy) > 0
            and isinstance(hierarchy[0], int)):
            hierarchy = [hierarchy]

        # Define thresholds for filling up the inside holes
        thresh_nuc, _ = cv2.threshold(nuclei_channel, 0, 255, cv2.THRESH_OTSU)
        med_excl = np.median(nuclei_channel[processed > 0])

        for contour, (_, _, _, parent) in zip(contours, hierarchy):

            # Only consider inside contours
            if parent < 0:
                continue

            # Create a mask for the current contour
            mask = np.zeros_like(processed)
            cv2.drawContours(mask, (contour,), -1, 255, -1)

            # If the hole in the fiber signal is below a nucleus, and the fiber
            # signal is still quite bright, then fill up the hole
            if (np.average(nuclei_channel[mask.nonzero()]) > thresh_nuc and
                np.average(fiber_channel[mask.nonzero()]) > med_excl):
                cv2.drawContours(processed, (contour,), -1, 255, -1)

        # Creating a boolean mask, later used to generate a masked array
        mask = processed.astype(bool)

        return mask

    @staticmethod
    def _get_nuclei_positions(labeled_image: np.ndarray,
                              mask: np.ndarray,
                              nuclei_channel: np.ndarray,
                              fiber_overlap_threshold: float,
                              minimum_nucleus_intensity: int,
                              maximum_nucleus_intensity: int,
                              minimum_nuclei_count: int
                              ) -> tuple[list[tuple[np.ndarray, np.ndarray]],
                                         list[tuple[np.ndarray, np.ndarray]]]:
        """Computes the center of the nuclei and determines whether they're
        positive or not.

        Args:
            labeled_image: The image containing the nuclei.
            mask: The boolean mask of the fibers.
            nuclei_channel: The channel of the image containing the nuclei.
            fiber_overlap_threshold: Fraction of area above which a nucleus is
                considered to be inside a fiber.
            minimum_nucleus_intensity: Any nucleus whose average brightness is
                lower than this value will be discarded.
            maximum_nucleus_intensity: Any nucleus whose average brightness is
                greater than this value will be discarded.
            minimum_nuclei_count: Nuclei located in a fiber containing less
                than this number of positive nuclei will be counted as
                negative.

        Returns:
            The list of the centers of nuclei outside of fibers, and the list
            of centers of nuclei inside fibers
        """

        nuclei_out_fiber = list()
        nuclei_in_fiber = list()

        # Detect the contours of the detected fibers
        contours, hierarchy = cv2.findContours(mask.astype(np.uint8),
                                               cv2.RETR_CCOMP,
                                               cv2.CHAIN_APPROX_SIMPLE)

        if hierarchy is None:
            return nuclei_out_fiber, nuclei_in_fiber

        hierarchy = np.squeeze(hierarchy).tolist()

        # If the hierarchy only contains one object, it is cast to int
        if isinstance(hierarchy, int):
            hierarchy = [hierarchy]
        if (isinstance(hierarchy, list)
            and len(hierarchy) > 0
            and isinstance(hierarchy[0], int)):
            hierarchy = [hierarchy]

        # Group the contours so that the holes are with their parents
        grouped = defaultdict(list)
        for i, (contour, (_, _, _, parent)) in enumerate(zip(contours,
                                                             hierarchy)):
            if parent < 0:
                grouped[i].append(contour)
            else:
                grouped[parent].append(contour)

        # Dict to keep track of the nuclei count in each fiber
        nuc_count = {i: 0 for i in grouped.keys()}

        # Create fiber masks to check in which fiber each positive nucleus is
        fib_masks = dict()
        for i, cont in grouped.items():
            fib_mask = np.zeros_like(mask, dtype=np.uint8)
            cv2.drawContours(fib_mask, cont, -1, 255, -1)
            fib_mask = fib_mask.astype(bool)
            fib_mask = ma.masked_array(labeled_image, fib_mask)
            fib_masks[i] = fib_mask

        # Masking the image containing the nuclei with the boolean mask of the
        # fibers
        masked_image = ma.masked_array(labeled_image, mask)

        # Each gray level on the nuclei image corresponds to one nucleus
        nb_nuc = np.max(labeled_image)

        # Buffers for storing the nuclei, their positivity status, and the
        # index of the fiber in which they are located
        nuclei = dict()
        positivity = dict()
        fiber_num = dict()

        # Iterating through the detected nuclei
        for i in range(1, nb_nuc + 1):

            # Getting a list of the detected pixels
            nucleus_y, nucleus_x = np.where(labeled_image == i)
            center_x, center_y = np.mean(nucleus_x), np.mean(nucleus_y)

            # Aborting for the current nucleus if it's not bright enough
            if not (minimum_nucleus_intensity <=
                    np.average(nuclei_channel[nucleus_y, nucleus_x]) <=
                    maximum_nucleus_intensity):
                nuclei[i] = None
                positivity[i] = None
                fiber_num[i] = None
                continue

            # Determining whether the nucleus is positive or negative
            in_fiber = ma.count_masked(masked_image[nucleus_y, nucleus_x])
            if in_fiber < fiber_overlap_threshold * nucleus_x.shape[0]:
                positive = False
            else:
                positive = True

            # Storing the nucleus position and positivity
            nuclei[i] = (center_x, center_y)
            positivity[i] = positive

            # Searching for the first fiber that contains part of the nucleus
            if positive:
                for j, fib_mask in fib_masks.items():
                    if ma.count_masked(fib_mask[nucleus_y, nucleus_x]):
                        nuc_count[j] += 1
                        fiber_num[i] = j
                        break
            else:
                fiber_num[i] = None

        # Excluding the fibers that contain too few nuclei
        fib_to_ban = tuple(i for i, count in nuc_count.items()
                           if count < minimum_nuclei_count)

        # Iterating again over the detected nuclei
        for i in range(1, nb_nuc + 1):

            # Means that the nucleus was skipped due to out-of-bounds intensity
            if nuclei[i] is None:
                continue

            center_x, center_y = nuclei[i]
            positive = positivity[i]
            fib_num = fiber_num[i]

            # Excluding the nuclei belonging to fibers with too few nuclei
            if fib_num is not None and fib_num in fib_to_ban:
                positive = False

            # Building the final lists of positive and negative nuclei
            if positive:
                nuclei_in_fiber.append((center_x, center_y))
            else:
                nuclei_out_fiber.append((center_x, center_y))

        return nuclei_out_fiber, nuclei_in_fiber
